import unittest
from unittest import expectedFailure
from unittest.mock import MagicMock, patch
import datetime as dt
import pandas as pd
import pytest

from meteole._arpege import ArpegeForecast
from meteole._arome import AromeForecast
from meteole.clients import MeteoFranceClient


class TestAromeForecast(unittest.TestCase):
    def setUp(self):
        self.precision = 0.01
        self.territory = "FRANCE"
        self._api_key = "fake_api_key"
        self.token = "fake_token"
        self.application_id = "fake_app_id"
        self.certs_path = "a/path/"
        self.client = MeteoFranceClient(
            api_key=self._api_key, token=self.token, application_id=self.application_id, certs_path=self.certs_path
        )

    @patch("meteole._arome.AromeForecast.get_capabilities")
    def test_initialization(self, mock_get_capabilities):
        mock_get_capabilities.return_value = None

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        self.assertEqual(forecast.precision, self.precision)
        self.assertEqual(forecast.territory, self.territory)
        self.assertEqual(forecast._client._api_key, self._api_key)
        self.assertEqual(forecast._client._token, self.token)
        self.assertEqual(forecast._client._application_id, self.application_id)
        self.assertEqual(forecast._client._verify, self.certs_path)
        self.assertEqual(forecast.MODEL_NAME, "arome")
        # mock_get_capabilities.assert_called_once()

    def test_invalid_precision(self):
        with self.assertRaises(ValueError):
            AromeForecast(self.client, precision=0.1)

    def test_invalid_territory(self):
        with self.assertRaises(ValueError):
            AromeForecast(self.client, territory="INVALID")

    @patch("meteole.clients.MeteoFranceClient.get")
    def test_get_capabilities(self, mock_get_request):
        mock_response = MagicMock()
        mock_response.text = """
            <wcs:Capabilities>
                <wcs:Contents>
                    <wcs:CoverageSummary>
                        <wcs:CoverageId>GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE___2024-10-31T00.00.00Z</wcs:CoverageId>
                        <ows:CoverageTitle>Geometric height</ows:CoverageTitle>
                        <wcs:CoverageSubtype>ReferenceableGridCoverage</wcs:CoverageSubtype>
                    </wcs:CoverageSummary>
                    <wcs:CoverageSummary>
                        <wcs:CoverageId>SHORT_WAVE_RADIATION_FLUX__GROUND_OR_WATER_SURFACE___2024-11-01T18.00.00Z_P2D</wcs:CoverageId>
                        <ows:CoverageTitle>short-wave radiation flux</ows:CoverageTitle>
                        <wcs:CoverageSubtype>ReferenceableGridCoverage</wcs:CoverageSubtype>
                    </wcs:CoverageSummary>
                </wcs:Contents>
            </wcs:Capabilities>
        """
        mock_get_request.return_value = mock_response

        arome = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )
        # arome.get_capabilities() is made during init
        # this means arome.capabilities exists now

        self.assertEqual(
            list(arome.capabilities["id"]),
            [
                "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE___2024-10-31T00.00.00Z",
                "SHORT_WAVE_RADIATION_FLUX__GROUND_OR_WATER_SURFACE___2024-11-01T18.00.00Z_P2D",
            ],
        )

    @patch("meteole.clients.MeteoFranceClient.get")
    def test_get_coverage_id(self, mock_get_request):
        mock_response = MagicMock()
        mock_response.text = """
            <wcs:Capabilities>
                <wcs:Contents>
                    <wcs:CoverageSummary>
                        <wcs:CoverageId>GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE___2024-10-31T00.00.00Z</wcs:CoverageId>
                        <ows:CoverageTitle>Geometric height</ows:CoverageTitle>
                        <wcs:CoverageSubtype>ReferenceableGridCoverage</wcs:CoverageSubtype>
                    </wcs:CoverageSummary>
                    <wcs:CoverageSummary>
                        <wcs:CoverageId>TOTAL_WATER_PRECIPITATION__GROUND_OR_WATER_SURFACE___2024-11-01T18.00.00Z_P2D</wcs:CoverageId>
                        <ows:CoverageTitle>short-wave radiation flux</ows:CoverageTitle>
                        <wcs:CoverageSubtype>ReferenceableGridCoverage</wcs:CoverageSubtype>
                    </wcs:CoverageSummary>
                </wcs:Contents>
            </wcs:Capabilities>
        """
        mock_get_request.return_value = mock_response

        arome = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )  # arome.capabilities is set up

        run = "2024-11-01T18.00.00Z"
        interval = "P2D"

        # test wrong indicator raises
        indicator = "wrong_indicator"

        with pytest.raises(ValueError):
            coverage_id = arome._get_coverage_id(indicator, run, interval)

        indicator = "TOTAL_WATER_PRECIPITATION__GROUND_OR_WATER_SURFACE"
        coverage_id = arome._get_coverage_id(indicator, run, interval)
        assert coverage_id == "TOTAL_WATER_PRECIPITATION__GROUND_OR_WATER_SURFACE___2024-11-01T18.00.00Z_P2D"

    @patch("meteole._arome.AromeForecast.get_capabilities")
    @patch("meteole.clients.MeteoFranceClient.get")
    def test_get_coverage_description(self, mock_get_request, mock_get_capabilities):
        mock_response = MagicMock()
        mock_response.text = """
            <wcs:CoverageDescriptions xmlns:wcs="http://www.opengis.net/wcs/2.0" xmlns:swe="http://www.opengis.net/swe/2.0" xmlns:gmlrgrid="http://www.opengis.net/gml/3.3/rgrid" xmlns:gmlcov="http://www.opengis.net/gmlcov/1.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:gml="http://www.opengis.net/gml/3.2" xsi:schemaLocation="     http://www.opengis.net/wcs/2.0 http://schemas.opengis.net/wcs/2.0/wcsAll.xsd     http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd     http://www.opengis.net/gmlcov/1.0 http://schemas.opengis.net/gmlcov/1.0/gmlcovAll.xsd     http://www.opengis.net/swe/2.0 http://schemas.opengis.net/sweCommon/2.0/swe.xsd     http://www.opengis.net/gml/3.3/rgrid http://schemas.opengis.net/gml/3.3/referenceableGrid.xsd">
            <wcs:CoverageDescription gml:id="U_COMPONENT_OF_WIND_GUST__SPECIFIC_HEIGHT_LEVEL_ABOVE_GROUND___2024-06-20T09.00.00Z">
                <wcs:CoverageId>U_COMPONENT_OF_WIND_GUST__SPECIFIC_HEIGHT_LEVEL_ABOVE_GROUND___2024-06-20T09.00.00Z</wcs:CoverageId>
                <wcs:ServiceParameters>
                <wcs:CoverageSubtype>ReferenceableGridCoverage</wcs:CoverageSubtype>
                <wcs:nativeFormat>application/wmo-grib</wcs:nativeFormat>
                </wcs:ServiceParameters>
            </wcs:CoverageDescription>
            </wcs:CoverageDescriptions>
        """

        mock_get_request.return_value = mock_response
        mock_get_capabilities.return_value = None

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        description = forecast._get_coverage_description("coverage_1", ensemble_number=None)
        self.assertIn("wcs:CoverageDescriptions", description)

    @patch("meteole._arome.AromeForecast.get_capabilities")
    @patch("meteole._arome.AromeForecast._grib_bytes_to_df")
    @patch("meteole._arome.AromeForecast._get_coverage_file")
    def test_get_data_single_forecast(self, mock_get_coverage_file, mock_grib_bytes_to_df, mock_get_capabilities):
        mock_grib_bytes_to_df.return_value = pd.DataFrame({"data": [1, 2, 3]})

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        df = forecast._get_data_single_forecast(
            coverage_id="coverage_1",
            height=None,
            pressure=None,
            ensemble_number=None,
            forecast_horizon=dt.timedelta(hours=0),
            lat=(37.5, 55.4),
            long=(-12, 16),
        )

        self.assertTrue("data" in df.columns)

    @patch("meteole._arome.AromeForecast.get_capabilities")
    @patch("meteole._arome.AromeForecast._grib_bytes_to_df")
    @patch("meteole._arome.AromeForecast._get_coverage_file")
    def test_get_data_single_forecast_with_height(
        self, mock_get_coverage_file, mock_grib_bytes_to_df, mock_get_capabilities
    ):
        mock_get_coverage_file.return_value = ""
        mock_grib_bytes_to_df.return_value = pd.DataFrame({"data": [1, 2, 3], "heightAboveGround": ["2", "2", "2"]})

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        df = forecast._get_data_single_forecast(
            coverage_id="coverage_1",
            height=2,
            pressure=None,
            ensemble_number=None,
            forecast_horizon=dt.timedelta(hours=0),
            lat=(37.5, 55.4),
            long=(-12, 16),
        )

        self.assertTrue("data_2m" in df.columns)

    def test_compute_closest_grid_points(self):
        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        coords = [48.8566, 50.0, 52.3, 2.35, 5.2561]
        expected_025 = [48.75, 50.0, 52.25, 2.25, 5.25]
        expected_01 = [48.9, 50.0, 52.3, 2.4, 5.3]
        expected_005 = [48.85, 50.0, 52.3, 2.35, 5.25]
        expected_001 = [48.86, 50.0, 52.3, 2.35, 5.26]
        expected_0025 = [48.85, 50.0, 52.3, 2.35, 5.250]

        for precision, expected_coords in zip(
            (0.25, 0.1, 0.05, 0.01, 0.025), (expected_025, expected_01, expected_005, expected_001, expected_0025)
        ):
            forecast.precision = precision
            for c, exp in zip(coords, expected_coords):
                self.assertEqual(forecast._compute_closest_grid_point(c), exp)

    @patch("meteole._arome.AromeForecast.get_coverage_description")
    @patch("meteole._arome.AromeForecast.get_capabilities")
    @patch("meteole._arome.AromeForecast._get_data_single_forecast")
    def test_get_coverage(self, mock_get_data_single_forecast, mock_get_capabilities, mock_get_coverage_description):
        mock_get_data_single_forecast.return_value = pd.DataFrame(
            {
                "latitude": [1, 2, 3],
                "longitude": [4, 5, 6],
                "time": [7, 8, 9],
                "step": [10, 11, 12],
                "valid_time": [16, 17, 18],
                "data": [19, 20, 21],  # this column name varies depending on the coverage_id
            }
        )
        mock_get_coverage_description.return_value = {
            "heights": [2],
            "forecast_horizons": [dt.timedelta(hours=0)],
            "pressures": [],
            "min_latitude": -90,
            "max_latitude": 90,
            "min_longitude": -90,
            "max_longitude": 90,
        }

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        forecast.get_coverage(
            coverage_id="toto",
            heights=[2],
            forecast_horizons=[dt.timedelta(hours=0)],
            lat=(37.5, 55.4),
            long=(-12, 16),
        )

        mock_get_data_single_forecast.assert_called_once_with(
            coverage_id="toto",
            height=2,
            pressure=None,
            ensemble_number=None,
            forecast_horizon=dt.timedelta(hours=0),
            lat=(37.5, 55.4),
            long=(-12, 16),
            temp_dir=None,
        )

    @patch("meteole._arome.AromeForecast.get_coverage_description")
    @patch("meteole._arome.AromeForecast.get_capabilities")
    @patch("meteole._arome.AromeForecast._get_data_single_forecast")
    def test_get_coverage_lat_lon(
        self, mock_get_data_single_forecast, mock_get_capabilities, mock_get_coverage_description
    ):
        """Tests the different ways that a user can provide lat and long to get_coverage (float and tuple)"""
        mock_get_data_single_forecast.return_value = pd.DataFrame(
            {
                "latitude": [1, 2, 3],
                "longitude": [4, 5, 6],
                "time": [7, 8, 9],
                "step": [10, 11, 12],
                "valid_time": [16, 17, 18],
                "data": [19, 20, 21],  # this column name varies depending on the coverage_id
            }
        )
        mock_get_coverage_description.return_value = {
            "heights": [],
            "forecast_horizons": [dt.timedelta(hours=0)],
            "pressures": [],
            "min_latitude": -90,
            "max_latitude": 90,
            "min_longitude": -90,
            "max_longitude": 90,
        }

        forecast = AromeForecast(
            self.client,
            precision=0.01,
            territory=self.territory,
        )

        for lat, expected_lat in zip((37.5689, (37.5689, 45.00986)), ((37.57, 37.57), (37.57, 45.01))):
            for long, expected_long in zip((2.568, (-1.566, 2.568)), ((2.57, 2.57), (-1.57, 2.57))):
                forecast.get_coverage(coverage_id="toto", lat=lat, long=long)
                mock_get_data_single_forecast.assert_called_once_with(
                    coverage_id="toto",
                    ensemble_number=None,
                    height=None,
                    pressure=None,
                    forecast_horizon=dt.timedelta(hours=0),
                    lat=expected_lat,
                    long=expected_long,
                    temp_dir=None,
                )
                mock_get_data_single_forecast.reset_mock()

    @patch("meteole._arome.AromeForecast.get_coverage_description")
    def test_get_forecast_horizons(self, mock_get_coverage_description):
        def side_effect(coverage_id):
            if coverage_id == "id1":
                return {
                    "forecast_horizons": [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2)],
                    "heights": [],
                    "pressures": [],
                }
            elif coverage_id == "id2":
                return {
                    "forecast_horizons": [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2)],
                    "heights": [],
                    "pressures": [],
                }

        mock_get_coverage_description.side_effect = side_effect

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        coverage_ids = ["id1", "id2"]
        expected_result = [
            [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2)],
            [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2)],
        ]
        result = forecast._get_forecast_horizons(coverage_ids)
        self.assertEqual(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_forecast_horizons")
    def test_find_common_forecast_horizons(self, mock_get_forecast_horizons):
        mock_get_forecast_horizons.return_value = [
            [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2), dt.timedelta(hours=3)],
            [dt.timedelta(hours=2), dt.timedelta(hours=3), dt.timedelta(hours=4), dt.timedelta(hours=5)],
            [dt.timedelta(hours=1), dt.timedelta(hours=2), dt.timedelta(hours=3), dt.timedelta(hours=6)],
        ]

        list_coverage_id = ["id1", "id2", "id3"]
        expected_result = [dt.timedelta(hours=2), dt.timedelta(hours=3)]

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )
        result = forecast.find_common_forecast_horizons(list_coverage_id)
        self.assertEqual(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_forecast_horizons")
    def test_validate_forecast_horizons_valid(self, mock_get_forecast_horizons):
        mock_get_forecast_horizons.return_value = [
            [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2), dt.timedelta(hours=3)],
            [dt.timedelta(hours=2), dt.timedelta(hours=3), dt.timedelta(hours=4), dt.timedelta(hours=5)],
        ]

        coverage_ids = ["id1", "id2"]
        forecast_horizons = [dt.timedelta(hours=2), dt.timedelta(hours=3)]
        expected_result = []

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )
        result = forecast._validate_forecast_horizons(coverage_ids, forecast_horizons)
        self.assertEqual(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_forecast_horizons")
    def test_validate_forecast_horizons_invalid(self, mock_get_forecast_horizons):
        mock_get_forecast_horizons.return_value = [
            [dt.timedelta(hours=0), dt.timedelta(hours=1), dt.timedelta(hours=2), dt.timedelta(hours=3)],
            [dt.timedelta(hours=2), dt.timedelta(hours=3), dt.timedelta(hours=4), dt.timedelta(hours=5)],
        ]

        coverage_ids = ["id1", "id2"]
        forecast_horizons = [dt.timedelta(hours=1), dt.timedelta(hours=2)]
        expected_result = ["id2"]

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )
        result = forecast._validate_forecast_horizons(coverage_ids, forecast_horizons)
        self.assertEqual(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_coverage_id")
    @patch("meteole._arome.AromeForecast.find_common_forecast_horizons")
    @patch("meteole._arome.AromeForecast._validate_forecast_horizons")
    @patch("meteole._arome.AromeForecast.get_coverage")
    def test_get_combined_coverage(
        self,
        mock_get_coverage,
        mock_validate_forecast_horizons,
        mock_find_common_forecast_horizons,
        mock_get_coverage_id,
    ):
        mock_get_coverage_id.side_effect = lambda indicator, run, interval: f"{indicator}_{run}_{interval}"
        mock_find_common_forecast_horizons.return_value = [dt.timedelta(hours=0)]
        mock_validate_forecast_horizons.return_value = []
        mock_get_coverage.side_effect = [
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data1": [10, 20],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data2": [30, 40],
                }
            ),
        ]

        indicator_names = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE",
            "BRIGHTNESS_TEMPERATURE__GROUND_OR_WATER_SURFACE",
        ]
        runs = ["2024-12-13T00.00.00Z"]
        heights = [None, 2]
        pressures = [None, None]
        intervals = ["", "P1D"]
        lat = (37.5, 55.4)
        long = (-12, 16)
        forecast_horizons = [dt.timedelta(hours=0)]

        expected_result = pd.DataFrame(
            {
                "latitude": [1, 2],
                "longitude": [3, 4],
                "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                "data1": [10, 20],
                "data2": [30, 40],
            }
        )

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        result = forecast.get_combined_coverage(
            indicator_names=indicator_names,
            runs=runs,
            heights=heights,
            pressures=pressures,
            intervals=intervals,
            lat=lat,
            long=long,
            forecast_horizons=forecast_horizons,
        )
        pd.testing.assert_frame_equal(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_coverage_id")
    @patch("meteole._arome.AromeForecast.find_common_forecast_horizons")
    @patch("meteole._arome.AromeForecast._validate_forecast_horizons")
    @patch("meteole._arome.AromeForecast.get_coverage")
    def test_get_combined_coverage_invalid_forecast_horizons(
        self,
        mock_get_coverage,
        mock_validate_forecast_horizons,
        mock_find_common_forecast_horizons,
        mock_get_coverage_id,
    ):
        mock_get_coverage_id.side_effect = lambda indicator, run, interval: f"{indicator}_{run}_{interval}"
        mock_find_common_forecast_horizons.return_value = [dt.timedelta(hours=0)]
        mock_validate_forecast_horizons.return_value = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE_2024-12-13T00.00.00Z"
        ]

        indicator_names = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE",
            "BRIGHTNESS_TEMPERATURE__GROUND_OR_WATER_SURFACE",
        ]
        runs = ["2024-12-13T00.00.00Z"]
        heights = [None, 2]
        pressures = [None, None]
        intervals = ["", "P1D"]
        lat = (37.5, 55.4)
        long = (-12, 16)
        forecast_horizons = [dt.timedelta(hours=0)]

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        with self.assertRaises(ValueError) as context:
            forecast.get_combined_coverage(
                indicator_names=indicator_names,
                runs=runs,
                heights=heights,
                pressures=pressures,
                intervals=intervals,
                lat=lat,
                long=long,
                forecast_horizons=forecast_horizons,
            )
        self.assertIn("are not valid for these coverage_ids", str(context.exception))

    @patch("meteole._arome.AromeForecast._get_coverage_id")
    @patch("meteole._arome.AromeForecast.find_common_forecast_horizons")
    @patch("meteole._arome.AromeForecast._validate_forecast_horizons")
    @patch("meteole._arome.AromeForecast.get_coverage")
    def test_get_combined_coverage_multiple_runs(
        self,
        mock_get_coverage,
        mock_validate_forecast_horizons,
        mock_find_common_forecast_horizons,
        mock_get_coverage_id,
    ):
        # Mock return values
        mock_get_coverage_id.side_effect = lambda indicator, run, interval: f"{indicator}_{run}_{interval}"
        mock_find_common_forecast_horizons.return_value = [dt.timedelta(hours=0)]
        mock_validate_forecast_horizons.return_value = []
        mock_get_coverage.side_effect = [
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data1": [10, 20],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data2": [30, 40],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-14T00.00.00Z", "2024-12-14T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data1": [100, 200],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-14T00.00.00Z", "2024-12-14T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data2": [300, 400],
                }
            ),
        ]

        indicator_names = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE",
            "BRIGHTNESS_TEMPERATURE__GROUND_OR_WATER_SURFACE",
        ]
        runs = ["2024-12-13T00.00.00Z", "2024-12-14T00.00.00Z"]
        heights = [None, 2]
        pressures = [None, None]
        intervals = ["", "P1D"]
        lat = (37.5, 55.4)
        long = (-12, 16)
        forecast_horizons = [dt.timedelta(hours=0)]

        expected_result = pd.DataFrame(
            {
                "latitude": [1, 2, 1, 2],
                "longitude": [3, 4, 3, 4],
                "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z", "2024-12-14T00.00.00Z", "2024-12-14T00.00.00Z"],
                "forecast_horizon": [
                    dt.timedelta(hours=0),
                    dt.timedelta(hours=0),
                    dt.timedelta(hours=0),
                    dt.timedelta(hours=0),
                ],
                "data1": [10, 20, 100, 200],
                "data2": [30, 40, 300, 400],
            }
        )

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        result = forecast.get_combined_coverage(
            indicator_names=indicator_names,
            runs=runs,
            heights=heights,
            pressures=pressures,
            intervals=intervals,
            lat=lat,
            long=long,
            forecast_horizons=forecast_horizons,
        )
        pd.testing.assert_frame_equal(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_coverage_id")
    @patch("meteole._arome.AromeForecast.find_common_forecast_horizons")
    @patch("meteole._arome.AromeForecast._validate_forecast_horizons")
    @patch("meteole._arome.AromeForecast.get_coverage")
    def test_get_combined_coverage_no_heights_or_pressures(
        self,
        mock_get_coverage,
        mock_validate_forecast_horizons,
        mock_find_common_forecast_horizons,
        mock_get_coverage_id,
    ):
        mock_get_coverage_id.side_effect = lambda indicator, run, interval: f"{indicator}_{run}_{interval}"
        mock_find_common_forecast_horizons.return_value = [dt.timedelta(hours=0)]
        mock_validate_forecast_horizons.return_value = []
        mock_get_coverage.side_effect = [
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data1": [10, 20],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data2": [30, 40],
                }
            ),
        ]

        indicator_names = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE",
            "BRIGHTNESS_TEMPERATURE__GROUND_OR_WATER_SURFACE",
        ]
        runs = ["2024-12-13T00.00.00Z"]
        heights = None
        pressures = None
        intervals = ["", "P1D"]
        lat = (37.5, 55.4)
        long = (-12, 16)
        forecast_horizons = [dt.timedelta(hours=0)]

        expected_result = pd.DataFrame(
            {
                "latitude": [1, 2],
                "longitude": [3, 4],
                "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                "data1": [10, 20],
                "data2": [30, 40],
            }
        )

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        result = forecast.get_combined_coverage(
            indicator_names, runs, heights, pressures, intervals, lat, long, forecast_horizons
        )
        pd.testing.assert_frame_equal(result, expected_result)

    @patch("meteole._arome.AromeForecast._get_coverage_id")
    @patch("meteole._arome.AromeForecast.find_common_forecast_horizons")
    @patch("meteole._arome.AromeForecast._validate_forecast_horizons")
    @patch("meteole._arome.AromeForecast.get_coverage")
    def test_get_combined_coverage_no_optional_params(
        self,
        mock_get_coverage,
        mock_validate_forecast_horizons,
        mock_find_common_forecast_horizons,
        mock_get_coverage_id,
    ):
        mock_get_coverage_id.side_effect = lambda indicator, run, interval: f"{indicator}_{run}_{interval}"
        mock_find_common_forecast_horizons.return_value = [0]
        mock_validate_forecast_horizons.return_value = []
        mock_get_coverage.side_effect = [
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data1": [10, 20],
                }
            ),
            pd.DataFrame(
                {
                    "latitude": [1, 2],
                    "longitude": [3, 4],
                    "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                    "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                    "data2": [30, 40],
                }
            ),
        ]

        indicator_names = [
            "GEOMETRIC_HEIGHT__GROUND_OR_WATER_SURFACE",
            "BRIGHTNESS_TEMPERATURE__GROUND_OR_WATER_SURFACE",
        ]
        runs = ["2024-12-13T00.00.00Z"]

        expected_result = pd.DataFrame(
            {
                "latitude": [1, 2],
                "longitude": [3, 4],
                "run": ["2024-12-13T00.00.00Z", "2024-12-13T00.00.00Z"],
                "forecast_horizon": [dt.timedelta(hours=0), dt.timedelta(hours=0)],
                "data1": [10, 20],
                "data2": [30, 40],
            }
        )

        forecast = AromeForecast(
            self.client,
            precision=self.precision,
            territory=self.territory,
        )

        result = forecast.get_combined_coverage(indicator_names, runs)
        pd.testing.assert_frame_equal(result, expected_result)


class TestArpegeForecast(unittest.TestCase):
    def setUp(self):
        self.territory = "EUROPE"
        self._api_key = "fake_api_key"
        self.token = "fake_token"
        self.application_id = "fake_app_id"
        self.client = MeteoFranceClient(token=self.token)

    @patch("meteole._arpege.ArpegeForecast.get_capabilities")
    def test_initialization(self, mock_get_capabilities):
        territory = "EUROPE"
        api_key = "test_api_key"
        token = "test_token"
        application_id = "test_app_id"

        client = MeteoFranceClient(api_key=api_key, token=token, application_id=application_id, certs_path="toto")

        arpege_forecast = ArpegeForecast(
            client,
            territory=territory,
        )

        self.assertEqual(arpege_forecast.territory, territory)
        self.assertEqual(arpege_forecast.precision, ArpegeForecast.RELATION_TERRITORY_TO_PREC_ARPEGE[territory])
        self.assertEqual(arpege_forecast._client._verify, "toto")
        self.assertEqual(arpege_forecast._client._api_key, api_key)
        self.assertEqual(arpege_forecast._client._token, token)
        self.assertEqual(arpege_forecast._client._application_id, application_id)
        self.assertEqual(arpege_forecast.MODEL_NAME, "arpege")
        # mock_get_capabilities.assert_called_once()

    @patch("meteole._arpege.ArpegeForecast.get_capabilities")
    def test_validate_parameters(self, mock_get_capabilities):
        valid_territory = "EUROPE"
        invalid_territory = "INVALID_TERRITORY"

        # Test with a valid territory
        client = MeteoFranceClient(api_key="toto")
        arpege_forecast = ArpegeForecast(client, territory=valid_territory)

        try:
            arpege_forecast._validate_parameters()
        except ValueError:
            self.fail("_validate_parameters raised ValueError unexpectedly with valid territory!")

        # Test with an invalid territory
        arpege_forecast.territory = invalid_territory
        with self.assertRaises(ValueError):
            arpege_forecast._validate_parameters()

    @patch("meteole._arpege.ArpegeForecast.get_capabilities")
    @patch("meteole.clients.MeteoFranceClient._connect")
    def test_entry_point(self, mock_MeteoFranceClient_connect, mock_get_capabilities):
        territory = "EUROPE"
        arpege_forecast = ArpegeForecast(self.client, territory=territory)
        expected_entry_point = f"wcs/MF-NWP-GLOBAL-ARPEGE-{ArpegeForecast.PRECISION_FLOAT_TO_STR[ArpegeForecast.RELATION_TERRITORY_TO_PREC_ARPEGE[territory]]}-{territory}-WCS"
        self.assertEqual(arpege_forecast._entry_point, expected_entry_point)


class TestGetAvailableFeature(unittest.TestCase):
    def setUp(self):
        self.grid_axis = [
            {
                "gmlrgrid:GeneralGridAxis": {
                    "gmlrgrid:gridAxesSpanned": "time",
                    "gmlrgrid:coefficients": "3600 7200 10800",
                }
            },
            {
                "gmlrgrid:GeneralGridAxis": {
                    "gmlrgrid:gridAxesSpanned": "height",
                    "gmlrgrid:coefficients": "100 200 300",
                }
            },
            {
                "gmlrgrid:GeneralGridAxis": {
                    "gmlrgrid:gridAxesSpanned": "pressure",
                    "gmlrgrid:coefficients": "1000 2000 3000",
                }
            },
        ]

    def test_get_available_feature_time(self):
        result = AromeForecast._get_available_feature(self.grid_axis, "time")
        self.assertEqual(result, [3600, 7200, 10800])

    def test_get_available_feature_height(self):
        result = AromeForecast._get_available_feature(self.grid_axis, "height")
        self.assertEqual(result, [100, 200, 300])

    def test_get_available_feature_pressure(self):
        result = AromeForecast._get_available_feature(self.grid_axis, "pressure")
        self.assertEqual(result, [1000, 2000, 3000])

    def test_get_available_feature_not_found(self):
        result = AromeForecast._get_available_feature(self.grid_axis, "nonexistent")
        self.assertEqual(result, [])



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any, final

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import pandas as pd
from requests import Response

from meteole.clients import BaseClient, MeteoFranceClient
from meteole.errors import MissingDataError

logger = logging.getLogger(__name__)


@final
class Vigilance:
    """Easy access to the vigilance data of Meteo-France.

    Resources are:
    - textesvigilance
    - cartevigilance

    Docs:
    - https://portail-api.meteofrance.fr/web/fr/api/DonneesPubliquesVigilance
    - https://donneespubliques.meteofrance.fr/client/document/descriptiftechnique_vigilancemetropole_
        donneespubliques_v4_20230911_307.pdf
    """

    API_VERSION: str = "v1"
    VIGILANCE_BASE_PATH: str = "DPVigilance/"
    PHENOMENO_IDS: dict[str, str] = {
        "1": "wind",
        "2": "rain",
        "3": "storm",
        "4": "flood",
        "5": "snow_ice",
        "6": "heat_wave",
        "7": "extreme_cold",
        "8": "avalanches",
        "9": "waves_submergence",
    }

    def __init__(
        self,
        client: BaseClient | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize attributes"""
        if client is not None:
            self._client = client
        else:
            # Try to instantiate it (can be user friendly)
            self._client = MeteoFranceClient(**kwargs)

    def get_bulletin(self) -> dict[str, Any]:
        """Retrieve the vigilance bulletin.

        Returns:
            Dictionary representing the vigilance bulletin
        """

        path: str = self.VIGILANCE_BASE_PATH + self.API_VERSION + "/textesvigilance/encours"
        logger.debug(f"GET {path}")

        try:
            resp: Response = self._client.get(path)
            return resp.json()

        except MissingDataError as e:
            if "no matching blob" in str(e):
                logger.warning("Ongoing vigilance requires no publication")
            else:
                logger.error(f"Unexpected error: {e}")
            return {}
        except Exception as e:
            logger.error(f"Unexpected error: {e}")
            return {}

    def get_map(self) -> dict[str, Any]:
        """Get the vigilance map with predicted risk displayed.

        Returns:
            Dictionary with the predicted risk.
        """
        path: str = self.VIGILANCE_BASE_PATH + self.API_VERSION + "/cartevigilance/encours"
        logger.debug(f"GET {path}")

        resp: Response = self._client.get(path)

        return resp.json()

    def get_phenomenon(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Get risk prediction by phenomenon and by domain.

        Returns:
            Tuple of:
                A DataFrame with phenomenon by id
                A DataFrame with phenomenon by domain
        """
        df_carte = pd.DataFrame(self.get_map())
        periods_data = df_carte.loc["periods", "product"]
        df_periods = pd.json_normalize(periods_data)

        df_j = df_periods[df_periods["echeance"] == "J"]
        df_j1 = df_periods[df_periods["echeance"] == "J1"]

        df_phenomenon_j = pd.json_normalize(df_j["per_phenomenon_items"].explode())
        df_phenomenon_j1 = pd.json_normalize(df_j1["per_phenomenon_items"].explode())
        df_phenomenon_j["echeance"] = "J"
        df_phenomenon_j1["echeance"] = "J1"

        df_phenomenon = pd.concat([df_phenomenon_j, df_phenomenon_j1]).reset_index(drop=True)
        df_phenomenon["phenomenon_libelle"] = df_phenomenon["phenomenon_id"].map(self.PHENOMENO_IDS)

        df_timelaps_j = pd.json_normalize(df_j["timelaps.domain_ids"].explode())
        df_timelaps_j1 = pd.json_normalize(df_j1["timelaps.domain_ids"].explode())
        df_timelaps_j["echeance"] = "J"
        df_timelaps_j1["echeance"] = "J1"

        df_timelaps = pd.concat([df_timelaps_j, df_timelaps_j1]).reset_index(drop=True)

        return df_phenomenon, df_timelaps

    def get_vignette(self) -> None:
        """Get png file."""
        path: str = self.VIGILANCE_BASE_PATH + self.API_VERSION + "/vignettenationale-J-et-J1/encours"

        logger.debug(f"GET {path}")
        resp: Response = self._client.get(path)

        if resp.status_code == 200:
            content: str | None = resp.headers.get("content-disposition")
            if content is not None:
                filename: str = content.split("filename=")[1]
                filename = filename.strip('"')

            with open(filename, "wb") as f:
                f.write(resp.content)

            img = mpimg.imread(BytesIO(resp.content), format="png")
            plt.imshow(img)
            plt.axis("off")
            plt.show()
        else:
            resp.raise_for_status()



from __future__ import annotations

import logging
from typing import Any, final

from meteole.clients import BaseClient, MeteoFranceClient
from meteole.forecast import WeatherForecast

logger = logging.getLogger(__name__)

AVAILABLE_ARPEGE_TERRITORY: list[str] = ["EUROPE", "GLOBE"]


@final
class ArpegePEForecast(WeatherForecast):
    """Access the ARPEGE numerical weather forecast data from Meteo-France API.

    Doc:
        - https://portail-api.meteofrance.fr/web/fr/api/pe-arpege

    Attributes:
        territory: Covered area (e.g., FRANCE, EUROPE, ...).
        precision: Precision value of the forecast.
        capabilities: DataFrame containing details on all available coverage ids.
    """

    MODEL_NAME: str = "pearpege"
    BASE_ENTRY_POINT: str = "wcs/MF-NWP-GLOBAL-PEARP"
    MODEL_TYPE: str = "ENSEMBLE"
    ENSEMBLE_NUMBERS: int = 35
    DEFAULT_TERRITORY: str = "EUROPE"
    RELATION_TERRITORY_TO_PREC_ARPEGE: dict[str, float] = {"EUROPE": 0.1, "GLOBE": 0.25}
    CLIENT_CLASS: type[BaseClient] = MeteoFranceClient

    def __init__(
        self,
        client: BaseClient | None = None,
        *,
        territory: str = "EUROPE",
        **kwargs: Any,
    ):
        """Initializes an ArpegeForecast object.

        The `precision` of the forecast is inferred from the specified `territory`.

        Args:
            territory: The ARPEGE territory to fetch. Defaults to "EUROPE".
            api_key: The API key for authentication. Defaults to None.
            token: The API token for authentication. Defaults to None.
            application_id: The Application ID for authentication. Defaults to None.

        Notes:
            - See `MeteoFranceClient` for additional details on the parameters `api_key`, `token`,
                and `application_id`.
            - Available territories are listed in the `AVAILABLE_TERRITORY` constant.
        """
        super().__init__(
            client=client,
            territory=territory,
            precision=self.RELATION_TERRITORY_TO_PREC_ARPEGE[territory],
            **kwargs,
        )

    def _validate_parameters(self) -> None:
        """Check the territory and the precision parameters.

        Raise:
            ValueError: At least, one parameter is not good.
        """
        if self.territory not in AVAILABLE_ARPEGE_TERRITORY:
            raise ValueError(f"The parameter precision must be in {AVAILABLE_ARPEGE_TERRITORY}")



from __future__ import annotations

import datetime as dt
import logging
import os
import re
import shutil
import tempfile
from abc import ABC, abstractmethod
from functools import reduce
from importlib.util import find_spec
from typing import Any
from warnings import warn

import pandas as pd
import xarray as xr
import xmltodict

from meteole.clients import BaseClient
from meteole.errors import MissingDataError

if find_spec("cfgrib") is None:
    raise ImportError(
        "The 'cfgrib' module is required to read Arome and Arpege GRIB files. Please install it using:\n\n"
        "  conda install -c conda-forge cfgrib\n\n"
    )

logger = logging.getLogger(__name__)


class WeatherForecast(ABC):
    """(Abstract)
    Base class for weather forecast models.

    Note: Currently, this class is highly related to Meteo-France models.
    This will not be the case in the future.

    Attributes:
        territory: Covered area (e.g., FRANCE, ANTIL, ...).
        precision: Precision value of the forecast.
        capabilities: DataFrame containing details on all available coverage ids.
    """

    # Class constants
    # Global
    API_VERSION: str = "1.0"
    PRECISION_FLOAT_TO_STR: dict[float, str] = {0.25: "025", 0.1: "01", 0.05: "005", 0.01: "001", 0.025: "0025"}
    FRANCE_METRO_LONGITUDES = (-5.1413, 9.5602)
    FRANCE_METRO_LATITUDES = (41.33356, 51.0889)

    # Model
    MODEL_NAME: str = "Defined in subclass"
    BASE_ENTRY_POINT: str = "Defined in subclass"
    MODEL_TYPE: str = "Defined in subclass"
    ENSEMBLE_NUMBERS: int = 1
    DEFAULT_TERRITORY: str = "FRANCE"
    DEFAULT_PRECISION: float = 0.01
    MAX_DECIMAL_PLACES: int = 4  # used to avoid floating point issues when finding the closest grid point
    CLIENT_CLASS: type[BaseClient]

    def __init__(
        self,
        client: BaseClient | None = None,
        *,
        territory: str = DEFAULT_TERRITORY,
        precision: float = DEFAULT_PRECISION,
        **kwargs: Any,
    ):
        """Initialize attributes.

        Args:
            territory: The ARPEGE territory to fetch.
            api_key: The API key for authentication. Defaults to None.
            token: The API token for authentication. Defaults to None.
            application_id: The Application ID for authentication. Defaults to None.
        """

        self.territory = territory  # "FRANCE", "ANTIL", or others (see API doc)
        self.precision = precision
        self._validate_parameters()

        self._capabilities: pd.DataFrame | None = None
        self._entry_point: str

        if self.MODEL_TYPE == "ENSEMBLE":
            self._entry_point = (
                f"{self.BASE_ENTRY_POINT}xxx-{self.PRECISION_FLOAT_TO_STR[self.precision]}-{self.territory}-WCS"
            )
        else:
            self._entry_point = (
                f"{self.BASE_ENTRY_POINT}-{self.PRECISION_FLOAT_TO_STR[self.precision]}-{self.territory}-WCS"
            )
        self._model_base_path = self.MODEL_NAME + "/" + self.API_VERSION

        if client is not None:
            self._client = client
        else:
            # Try to instantiate it (can be user friendly)
            self._client = self.CLIENT_CLASS(**kwargs)

    @property
    def indicators(self) -> list[str]:
        """Computes the list of all indicators from self.capabilities

        Returns: List of all indicators
        """
        return self.capabilities["indicator"].unique().tolist()

    @property
    def instant_indicators(self) -> list[str]:
        """Computes the list of instant indicators from self.capabilities

        Returns: List of instant indicators
        """
        return self.capabilities[self.capabilities["interval"] == ""]["indicator"].unique().tolist()

    @property
    def INDICATORS(self) -> list[str]:
        """Deprecated, will be removed in a future version, use indicators instead

        Returns: List of all indicators
        """
        warn(
            ("Deprecated, will be removed in a future version, use indicators instead"),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.indicators

    @property
    def INSTANT_INDICATORS(self) -> list[str]:
        """Deprecated, will be removed in a future version, use instant_indicators instead

        Returns: List of instant indicators
        """
        warn(
            ("Deprecated, will be removed in a future version, use instant_indicators instead"),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.instant_indicators

    @property
    def capabilities(self) -> pd.DataFrame:
        """Getter method of the capabilities attribute.

        Returns:
            DataFrame of details on all available coverage ids.
        """
        if self._capabilities is None:
            self._capabilities = self._build_capabilities()
        return self._capabilities

    @abstractmethod
    def _validate_parameters(self) -> None:
        """Check the territory and the precision parameters.

        Raise:
            ValueError: At least, one parameter is not good.
        """
        raise NotImplementedError

    def get_capabilities(self) -> pd.DataFrame:
        """Explicit "getter method" of the capabilities attribute.

        Returns:
            DataFrame of details on all available coverage ids.
        """
        return self.capabilities

    def get_coverage_description(
        self, coverage_id: str, ensemble_numbers: list[int | None] | None = None
    ) -> dict[str, Any]:
        """Return the available axis (times, heights) of a coverage.

        TODO: Other informations can be fetched, not yet implemented.

        Args:
            coverage_id: An id of a coverage, use get_capabilities() to get them.
            ensemble_numbers: For ensemble models only, numbers of the desired
                   ensemble members. If None, defaults to the member 0.
        Returns:
            A dictionary containing more info on the coverage.
        """
        numbers_to_fetch: list[int | None]

        if self.MODEL_TYPE == "ENSEMBLE":
            if ensemble_numbers is None:
                numbers_to_fetch = [0]
            else:
                numbers_to_fetch = ensemble_numbers
            coverage_description = {}
        else:
            numbers_to_fetch = [None]

        for ensemble_number in numbers_to_fetch:
            description = self._get_coverage_description(coverage_id, ensemble_number)
            grid_axis = description["wcs:CoverageDescriptions"]["wcs:CoverageDescription"]["gml:domainSet"][
                "gmlrgrid:ReferenceableGridByVectors"
            ]["gmlrgrid:generalGridAxis"]

            coverage_description_single = {
                "forecast_horizons": [
                    dt.timedelta(seconds=time) for time in self._get_available_feature(grid_axis, "time")
                ],
                "heights": self._get_available_feature(grid_axis, "height"),
                "pressures": self._get_available_feature(grid_axis, "pressure"),
            }

            # Find min and max latitude and longitude
            envelope = description["wcs:CoverageDescriptions"]["wcs:CoverageDescription"]["gml:boundedBy"][
                "gml:EnvelopeWithTimePeriod"
            ]
            lower = envelope["gml:lowerCorner"]  # '-12 37.5'
            upper = envelope["gml:upperCorner"]  # '16 55.4'
            lower_vals = [float(val) for val in lower.split()]
            upper_vals = [float(val) for val in upper.split()]

            axis_labels = envelope.get("@axisLabels")
            if axis_labels and "long" in axis_labels and "lat" in axis_labels:
                labels = axis_labels.split()
                idx_long = labels.index("long")
                idx_lat = labels.index("lat")

                lower_long = lower_vals[idx_long]
                lower_lat = lower_vals[idx_lat]
                upper_long = upper_vals[idx_long]
                upper_lat = upper_vals[idx_lat]
            else:
                lower_long, lower_lat = lower_vals[:2]
                upper_long, upper_lat = upper_vals[:2]

            coverage_description_single["min_latitude"] = lower_lat
            coverage_description_single["max_latitude"] = upper_lat
            coverage_description_single["min_longitude"] = lower_long
            coverage_description_single["max_longitude"] = upper_long

            if ensemble_number is None or len(numbers_to_fetch) == 1:
                coverage_description = coverage_description_single
            else:
                coverage_description[f"number_{ensemble_number}"] = coverage_description_single

        return coverage_description

    def get_coverage(
        self,
        indicator: str | None = None,
        lat: tuple | float = FRANCE_METRO_LATITUDES,
        long: tuple | float = FRANCE_METRO_LONGITUDES,
        ensemble_numbers: list[int] | None = None,
        heights: list[int] | None = None,
        pressures: list[int] | None = None,
        forecast_horizons: list[dt.timedelta] | None = None,
        run: str | None = None,
        interval: str | None = None,
        coverage_id: str = "",
        temp_dir: str | None = None,
    ) -> pd.DataFrame:
        """Return the coverage data (i.e., the weather forecast data).

        Args:
            indicator: Indicator of a coverage to retrieve.
            lat (long): Minimum and maximum latitude (longitude), or latitude (longitude) of the desired location.
                        The closest grid point to the requested coordinate will be used.
            ensemble_numbers: For ensemble models only, numbers of the desired
                   ensemble members. If None, defaults to the member 0.
            heights: Heights in meters.
            pressures: Pressures in hPa.
            forecast_horizons: List of timedelta, representing the forecast horizons in hours.
            run: The model inference timestamp. If None, defaults to the latest available run.
                Expected format: "YYYY-MM-DDTHH:MM:SSZ".
            interval: The aggregation period. Must be None for instant indicators;
                    raises an error if specified. Defaults to "P1D" for time-aggregated indicators such
                    as TOTAL_PRECIPITATION.
            coverage_id: An id of a coverage, use get_capabilities() to get them.
            temp_dir (str | None): Directory to store the temporary file. Defaults to None.

        Returns:
            pd.DataFrame: The complete run for the specified execution.
        """
        # Numbers cannot be None if the model type is ENSEMBLE
        if self.MODEL_TYPE == "ENSEMBLE":
            if ensemble_numbers is None:
                ensemble_numbers = [0]
            logger.info(f"Using {len(ensemble_numbers)} ensemble members")

        # Ensure we only have one of coverage_id, indicator
        if not bool(indicator) ^ bool(coverage_id):
            raise ValueError("Argument `indicator` or `coverage_id` need to be set (only one of them)")
        if indicator is not None:
            coverage_id = self._get_coverage_id(indicator, run, interval)

        logger.info(f"Using `coverage_id={coverage_id}`")

        axis = self.get_coverage_description(coverage_id)

        # Handle lat,long inputs (needs axis to check bounds)
        user_lat, user_long = lat, long
        lat, long = self._check_and_format_coords(lat, long, axis)
        logger.info(f"Using `lat={lat} (user input: {user_lat})`")
        logger.info(f"Using `long={long} (user input: {user_long})`")

        heights = self._raise_if_invalid_or_fetch_default("heights", heights, axis["heights"])
        pressures = self._raise_if_invalid_or_fetch_default("pressures", pressures, axis["pressures"])
        forecast_horizons = self._raise_if_invalid_or_fetch_default(
            "forecast_horizons", forecast_horizons, axis["forecast_horizons"]
        )

        df_list = [
            self._get_data_single_forecast(
                coverage_id=coverage_id,
                ensemble_number=ensemble_number,
                height=height if height != -1 else None,
                pressure=pressure if pressure != -1 else None,
                forecast_horizon=forecast_horizon,
                lat=lat,
                long=long,
                temp_dir=temp_dir,
            )
            for forecast_horizon in forecast_horizons
            for pressure in pressures
            for height in heights
            for ensemble_number in ([None] if (ensemble_numbers is None) else ensemble_numbers)
        ]

        return pd.concat(df_list, axis=0).reset_index(drop=True)

    def _check_and_format_coords(
        self, lat: float | tuple[float, float], long: float | tuple[float, float], axis: dict[str, Any]
    ) -> tuple[tuple[float, float], tuple[float, float]]:
        """Formats lat, long arguments passed to get_coverage:
            - Rounds all coordinates to the closest grid point
            - If a single float is passed, converts it to a tuple (value,value)

        Args:
            lat, long : tuple (min,max) or float.
            lat (long): Minimum and maximum latitude (longitude), or latitude (longitude) of the desired location.
                    The closest grid point to the requested coordinate will be used.

        Returns:
            formatted coordinates: (min_lat, max_lat), (min_long, max_long)
        """
        if isinstance(lat, (int, float)):
            min_lat, max_lat = lat, lat
        else:
            min_lat, max_lat = lat
        if isinstance(long, (int, float)):
            min_long, max_long = long, long
        else:
            min_long, max_long = long
        min_long, max_long = self._compute_closest_grid_point(min_long), self._compute_closest_grid_point(max_long)
        min_lat, max_lat = self._compute_closest_grid_point(min_lat), self._compute_closest_grid_point(max_lat)
        if min_lat < axis["min_latitude"]:
            raise ValueError(f"Lower latitude is out of bounds (must be >= {axis['min_latitude']}).")
        if max_lat > axis["max_latitude"]:
            raise ValueError(f"Upper latitude is out of bounds (must be <= {axis['max_latitude']}).")
        if min_long < axis["min_longitude"]:
            raise ValueError(f"Lower longitude is out of bounds (must be >= {axis['min_longitude']}).")
        if max_long > axis["max_longitude"]:
            raise ValueError(f"Upper longitude is out of bounds (must be <= {axis['max_longitude']}).")
        return (min_lat, max_lat), (min_long, max_long)

    def _compute_closest_grid_point(self, coord: float) -> float:
        """Returns the coordinate of the closest grid point"""

        # The grid points are regularly spaced at intervals of 'precision'
        coord_grid = round(coord / self.precision) * self.precision
        coord_grid = round(coord_grid, self.MAX_DECIMAL_PLACES)  # avoid floating point issues
        return coord_grid

    def _build_capabilities(self) -> pd.DataFrame:
        """(Protected)
        Fetch and build the model capabilities.

        Returns:
            DataFrame all the details.
        """

        logger.info("Fetching all available coverages...")

        capabilities = self._fetch_capabilities()
        df_capabilities = pd.DataFrame(capabilities["wcs:Capabilities"]["wcs:Contents"]["wcs:CoverageSummary"])
        df_capabilities = df_capabilities.rename(
            columns={
                "wcs:CoverageId": "id",
                "ows:Title": "title",
                "wcs:CoverageSubtype": "subtype",
            }
        )
        df_capabilities["indicator"] = [coverage_id.split("___")[0] for coverage_id in df_capabilities["id"]]
        df_capabilities["run"] = [
            coverage_id.split("___")[1].split("Z")[0] + "Z" for coverage_id in df_capabilities["id"]
        ]
        df_capabilities["interval"] = [
            coverage_id.split("___")[1].split("Z")[1].strip("_") for coverage_id in df_capabilities["id"]
        ]

        nb_indicators = len(df_capabilities["indicator"].unique())
        nb_coverage_ids = df_capabilities.shape[0]
        runs = df_capabilities["run"].unique()

        logger.info(
            f"\n"
            f"\t Successfully fetched {nb_coverage_ids} coverages,\n"
            f"\t representing {nb_indicators} different indicators,\n"
            f"\t across the last {len(runs)} runs (from {runs.min()} to {runs.max()})\n"
            f"\n"
            f"\t Default run for `get_coverage`: {runs.max()})"
        )

        return df_capabilities

    def _get_coverage_id(
        self,
        indicator: str,
        run: str | None = None,
        interval: str | None = None,
    ) -> str:
        """(Protected)
        Retrieve a `coverage_id` from the capabilities based on the provided parameters.

        Args:
            indicator: The indicator to retrieve. This parameter is required.
            run: The model inference timestamp. If None, defaults to the latest available run.
                Expected format: "YYYY-MM-DDTHH:MM:SSZ". Defaults to None.
            interval: The aggregation period. Must be None for instant indicators;
                raises an error if specified. Defaults to "P1D" for time-aggregated indicators such as
                TOTAL_PRECIPITATION.

        Returns:
            str: The `coverage_id` corresponding to the given parameters.

        Raises:
            ValueError: If `indicator` is missing or invalid.
            ValueError: If `interval` is invalid or required but missing.
        """
        capabilities = self.capabilities[self.capabilities["indicator"] == indicator]

        if indicator not in self.indicators:
            raise ValueError(f"Unknown `indicator` - checkout `{self.MODEL_NAME}.indicators` to have the full list.")

        if run is None:
            run = capabilities.run.max()
            logger.info(f"Using latest `run={run}`.")

        try:
            dt.datetime.strptime(run, "%Y-%m-%dT%H.%M.%SZ")
        except ValueError as exc:
            raise ValueError(f"Run '{run}' is invalid. Expected format 'YYYY-MM-DDTHH.MM.SSZ'") from exc

        valid_runs = capabilities["run"].unique().tolist()
        if run not in valid_runs:
            raise ValueError(f"Run '{run}' is invalid. Valid runs : {valid_runs}")

        # handle interval
        valid_intervals = capabilities["interval"].unique().tolist()

        if indicator in self.instant_indicators:
            if not interval:
                # no interval is expected for instant indicators
                pass
            else:
                raise ValueError(
                    f"interval={interval} is invalid. No interval is expected (=set to None) for instant "
                    "indicator `{indicator}`."
                )
        else:
            if not interval:
                interval = valid_intervals[0]
                logger.info(
                    f"`interval=None` is invalid  for non-instant indicators. Using default `interval={interval}`"
                )
            elif interval not in valid_intervals:
                raise ValueError(
                    f"interval={interval} is invalid  for non-instant indicators. `{indicator}`."
                    f" Use valid intervals: {valid_intervals}"
                )

        coverage_id = f"{indicator}___{run}"

        if interval:
            coverage_id += f"_{interval}"

        return coverage_id

    def _raise_if_invalid_or_fetch_default(self, param_name: str, inputs: list | None, availables: list) -> list:
        """(Protected)
        Checks validity of `inputs`.

        Checks if the elements in `inputs` are in `availables` and raises a ValueError if not.
        If `inputs` is empty or None, uses the first element from `availables` as the default value.

        Args:
            param_name (str): The name of the parameter to validate.
            inputs (list[int] | None): The list of inputs to validate.
            availables (list[int]): The list of available values.

        Returns:
            list[int]: The validated list of inputs or the default value.

        Raises:
            ValueError: If any of the inputs are not in `availables`.
        """
        if inputs:
            for input_value in inputs:
                if input_value not in availables:
                    raise ValueError(f"`{param_name}={inputs}` is invalid. Available {param_name}: {availables}")
        else:
            inputs = availables[:1] or [
                -1
            ]  # using [-1] make sure we have an iterable. Using None makes things too complicated with mypy...
            if inputs[0] != -1:
                logger.info(f"Using `{param_name}={inputs}`")
        return inputs

    def _fetch_capabilities(self) -> dict[Any, Any]:
        """Fetch the model capabilities.

        Returns:
            Raw capabilities (dictionary).
        """

        if self.MODEL_TYPE == "ENSEMBLE":
            url = f"{self._model_base_path}/{self._entry_point.replace('xxx', '001')}/GetCapabilities"
        else:
            url = f"{self._model_base_path}/{self._entry_point}/GetCapabilities"

        params = {
            "service": "WCS",
            "version": "2.0.1",
            "language": "eng",
        }
        try:
            response = self._client.get(url, params=params)
        except MissingDataError as e:
            logger.error(f"Error fetching the capabilities: {e}")
            logger.error(f"URL: {url}")
            logger.error(f"Params: {params}")
            raise e

        xml = response.text

        try:
            return xmltodict.parse(xml)
        except MissingDataError as e:
            logger.error(f"Error parsing the XML response: {e}")
            logger.error(f"Response: {xml}")
            raise e

    def _get_coverage_description(self, coverage_id: str, ensemble_number: int | None) -> dict[Any, Any]:
        """(Protected)
        Get the description of a coverage.

        Warning:
            The return value is the raw XML data.
            Not yet parsed to be usable.
            In the future, it should be possible to use it to
            get the available heights, times, latitudes and longitudes of the forecast.

        Args:
            coverage_id (str): the Coverage ID. Use :meth:`get_coverage` to access the available coverage ids.
                By default use the latest temperature coverage ID.
            ensemble_number: For ensemble models only, number of the desired ensemble member.

        Returns:
            description (dict): the description of the coverage.
        """
        if self.MODEL_TYPE == "ENSEMBLE":
            url = (
                f"{self._model_base_path}/{self._entry_point.replace('xxx', f'{ensemble_number:03}')}/DescribeCoverage"
            )
        else:
            url = f"{self._model_base_path}/{self._entry_point}/DescribeCoverage"
        params = {
            "service": "WCS",
            "version": "2.0.1",
            "coverageid": coverage_id,
        }
        response = self._client.get(url, params=params)
        return xmltodict.parse(response.text)

    def _grib_bytes_to_df(
        self,
        grib_str: bytes,
        temp_dir: str | None = None,
    ) -> pd.DataFrame:
        """(Protected)
        Converts GRIB data (in binary format) into a pandas DataFrame.

        This method writes the binary GRIB data to a temporary file, reads it using
        the `cfgrib` engine via xarray, and converts the resulting xarray Dataset
        into a pandas DataFrame.

        Args:
            grib_str (bytes): Binary GRIB data as a byte string.
            temp_dir (str | None): Directory to store the temporary file. Defaults to None.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the extracted GRIB data,
            with columns like `time`, `latitude`, `longitude`, and any associated
            variables from the GRIB file.

        Raises:
            ValueError: If the input `grib_str` is not of type `bytes` or `bytearray`.

        Notes:
            - The method requires the `cfgrib` engine to be installed.
            - The temporary file used for parsing is automatically deleted after use.
            - Ensure the input GRIB data is valid and encoded in a binary format.
        """
        created_temp_dir = False

        if temp_dir:
            if not os.path.exists(temp_dir):
                os.makedirs(temp_dir)
                created_temp_dir = True
            temp_subdir = os.path.join(temp_dir, "temp_grib")
            os.makedirs(temp_subdir, exist_ok=True)
        else:
            temp_subdir = tempfile.mkdtemp()

        with tempfile.NamedTemporaryFile(dir=temp_subdir, delete=False) as temp_file:
            # Write the GRIB binary data to the temporary file
            temp_file.write(grib_str)
            temp_file.flush()  # Ensure the data is written to disk

            # Open the GRIB file as an xarray Dataset using the cfgrib engine
            ds = xr.open_dataset(temp_file.name, engine="cfgrib")

            # Convert the Dataset to a pandas DataFrame
            df = ds.to_dataframe().reset_index()

        if created_temp_dir and temp_dir is not None:
            shutil.rmtree(temp_dir)
        else:
            shutil.rmtree(temp_subdir)

        return df

    def _get_data_single_forecast(
        self,
        coverage_id: str,
        forecast_horizon: dt.timedelta,
        ensemble_number: int | None,
        pressure: int | None,
        height: int | None,
        lat: tuple,
        long: tuple,
        temp_dir: str | None = None,
    ) -> pd.DataFrame:
        """(Protected)
        Return the forecast's data for a given time and indicator.

        Args:
            coverage_id (str): the indicator.
            height (int): height in meters
            pressure (int): pressure in hPa
            forecast_horizon (dt.timedelta): the forecast horizon (how much time ahead?)
            ensemble_number (int): For ensemble models only, number of the desired ensemble member.
            lat (tuple): minimum and maximum latitude
            long (tuple): minimum and maximum longitude
            temp_dir (str | None): Directory to store the temporary file. Defaults to None.

        Returns:
            pd.DataFrame: The forecast for the specified time.
        """

        grib_binary: bytes = self._get_coverage_file(
            coverage_id=coverage_id,
            ensemble_number=ensemble_number,
            height=height,
            pressure=pressure,
            forecast_horizon_in_seconds=int(forecast_horizon.total_seconds()),
            lat=lat,
            long=long,
        )

        df: pd.DataFrame = self._grib_bytes_to_df(grib_binary, temp_dir=temp_dir)

        if self.MODEL_NAME == "pearpege":
            # for unclear reasons, the pearpege API does not accept lat, long
            # parameters unlike the other models API.
            # So we retrieve all the domain and then filter the results
            # for the desired lat, long in _get_data_single_forecast
            df = df.loc[
                (df["latitude"] <= lat[1])
                & (df["latitude"] >= lat[0])
                & (df["longitude"] <= long[1])
                & (df["longitude"] >= long[0])
            ]
        # Drop and rename columns
        df.drop(columns=["surface", "valid_time"], errors="ignore", inplace=True)
        df.rename(
            columns={
                "time": "run",
                "step": "forecast_horizon",
            },
            inplace=True,
        )
        if ensemble_number is None:
            known_columns = {"latitude", "longitude", "run", "forecast_horizon", "heightAboveGround", "isobaricInhPa"}
        else:
            known_columns = {
                "latitude",
                "longitude",
                "number",
                "run",
                "forecast_horizon",
                "heightAboveGround",
                "isobaricInhPa",
            }

        indicator_column = (set(df.columns) - known_columns).pop()

        if indicator_column == "unknown":
            base_name = "".join([word[0] for word in coverage_id.split("__")[0].split("_")]).lower()
        else:
            base_name = re.sub(r"\d.*", "", indicator_column)

        if "heightAboveGround" in df.columns:
            suffix = f"_{int(df['heightAboveGround'].iloc[0])}m"
        elif "isobaricInhPa" in df.columns:
            suffix = f"_{int(df['isobaricInhPa'].iloc[0])}hpa"
        else:
            suffix = ""

        new_indicator_column = f"{base_name}{suffix}"
        df.rename(columns={indicator_column: new_indicator_column}, inplace=True)
        if self.MODEL_TYPE == "ENSEMBLE":
            df.rename(columns={"number": "ensemble_number"}, inplace=True)

        df.drop(
            columns=["isobaricInhPa", "heightAboveGround", "meanSea", "potentialVorticity"],
            errors="ignore",
            inplace=True,
        )

        return df

    def _get_coverage_file(
        self,
        coverage_id: str,
        ensemble_number: int | None,
        height: int | None = None,
        pressure: int | None = None,
        forecast_horizon_in_seconds: int = 0,
        lat: tuple = (37.5, 55.4),
        long: tuple = (-12, 16),
    ) -> bytes:
        """(Protected)
        Retrieves data for a specified model prediction.

        Args:
            coverage_id (str): The coverage ID to retrieve. Use `get_coverage` to list available coverage IDs.
            ensemble_number: For ensemble models only, number of the desired ensemble member.
            height (int, optional): The height above ground level in meters. Defaults to 2 meters.
                If not provided, no height subset is applied.
            pressure (int, optional): The pressure level in hPa. If not provided, no pressure subset is applied.
            forecast_horizon_in_seconds (int, optional): The forecast horizon in seconds into the future.
                Defaults to 0 (current time).
            lat (tuple[float, float], optional): Tuple specifying the minimum and maximum latitudes.
                Defaults to (37.5, 55.4), covering the latitudes of France.
            long (tuple[float, float], optional): Tuple specifying the minimum and maximum longitudes.
                Defaults to (-12, 16), covering the longitudes of France.

        Returns:
            Path: The file path to the saved raster data.

        Notes:
            - If the file does not exist in the cache, it will be fetched from the API and saved.
            - Supported subsets include pressure, height, time, latitude, and longitude.

        See Also:
            raster.plot_tiff_file: Method for plotting raster data stored in TIFF format.
        """
        if ensemble_number is None:
            url = f"{self._model_base_path}/{self._entry_point}/GetCoverage"
        else:
            url = f"{self._model_base_path}/{self._entry_point.replace('xxx', f'{ensemble_number:03}')}/GetCoverage"

        if self.MODEL_NAME == "pearpege":
            # for unclear reasons, the pearpege API does not accept lat, long
            # parameters unlike the other models API.
            # So we retrieve all the domain and then filter the results
            # for the desired lat, long in _get_data_single_forecast
            subset = [
                *([f"pressure({pressure})"] if pressure is not None else []),
                *([f"height({height})"] if height is not None else []),
                f"time({forecast_horizon_in_seconds})",
            ]
        else:
            subset = [
                *([f"pressure({pressure})"] if pressure is not None else []),
                *([f"height({height})"] if height is not None else []),
                f"time({forecast_horizon_in_seconds})",
                f"lat({lat[0]},{lat[1]})",
                f"long({long[0]},{long[1]})",
            ]

        params = {
            "service": "WCS",
            "version": "2.0.1",
            "coverageid": coverage_id,
            "format": "application/wmo-grib",
            "subset": subset,
        }

        response = self._client.get(url, params=params)

        return response.content

    @staticmethod
    def _get_available_feature(grid_axis: list[dict[str, Any]], feature_name: str) -> list[int]:
        """(Protected)
        Retrieve available features.

        Args:
            grid_axis (list[dict[str, Any]]): A list of dictionaries where each dictionary represents a grid axis. Each dictionary contains
            information about the axis, such as the grid axes it spans and the associated coefficients.

            feature_name (str): Name of the feature you want to retrieve to filter the grid axes to find those that match the feature.

        Returns:
            List of available feature.
        """
        feature_grid_axis: list[dict[str, Any]] = [
            ax for ax in grid_axis if str(ax["gmlrgrid:GeneralGridAxis"]["gmlrgrid:gridAxesSpanned"]) == feature_name
        ]

        features: list[int] = (
            list(map(int, feature_grid_axis[0]["gmlrgrid:GeneralGridAxis"]["gmlrgrid:coefficients"].split(" ")))
            if len(feature_grid_axis) > 0
            else []
        )
        return features

    def get_combined_coverage(
        self,
        indicator_names: list[str],
        runs: list[str | None] | None = None,
        ensemble_numbers: list[int] | None = None,
        heights: list[int] | None = None,
        pressures: list[int] | None = None,
        intervals: list[str | None] | None = None,
        lat: tuple = FRANCE_METRO_LATITUDES,
        long: tuple = FRANCE_METRO_LONGITUDES,
        forecast_horizons: list[dt.timedelta] | None = None,
        temp_dir: str | None = None,
    ) -> pd.DataFrame:
        """
        Get a combined DataFrame of coverage data for multiple indicators and different runs.

        This method retrieves and aggregates coverage data for specified indicators, with options
        to filter by height, pressure, and forecast_horizon. It returns a concatenated DataFrame
        containing the coverage data for all provided runs.

        Args:
            indicator_names (list[str]): A list of indicator names to retrieve data for.
            runs (list[str]): A list of runs for each indicator. Format should be "YYYY-MM-DDTHH:MM:SSZ".
            ensemble_numbers: For ensemble models only, numbers of the desired
                   ensemble members. If None, defaults to the member 0.
            heights (list[int] | None): A list of heights in meters to filter by (default is None).
            pressures (list[int] | None): A list of pressures in hPa to filter by (default is None).
            intervals (list[str] | None): A list of aggregation periods (default is None).
                    Must be `None` or "" for instant indicators ; otherwise, raises an exception.
                    Defaults to 'P1D' for time-aggregated indicators.
            lat (tuple): The latitude range as (min_latitude, max_latitude). Defaults to FRANCE_METRO_LATITUDES.
            long (tuple): The longitude range as (min_longitude, max_longitude). Defaults to FRANCE_METRO_LONGITUDES.
            forecast_horizons (list[dt.timedelta] | None): A list of forecast horizon values in dt.timedelta. Defaults to None.
            temp_dir (str | None): Directory to store the temporary file. Defaults to None.

        Returns:
            pd.DataFrame: A combined DataFrame containing coverage data for all specified runs and indicators.

        Raises:
            ValueError: If the length of `heights` does not match the length of `indicator_names`.
        """
        # Numbers cannot be None if the model type is ENSEMBLE
        if self.MODEL_TYPE == "ENSEMBLE" and ensemble_numbers is None:
            ensemble_numbers = [0]

        if runs is None:
            runs = [None]
        coverages = [
            self._get_combined_coverage_for_single_run(
                indicator_names=indicator_names,
                run=run,
                lat=lat,
                long=long,
                ensemble_numbers=ensemble_numbers,
                heights=heights,
                pressures=pressures,
                intervals=intervals,
                forecast_horizons=forecast_horizons,
                temp_dir=temp_dir,
            )
            for run in runs
        ]
        return pd.concat(coverages, axis=0).reset_index(drop=True)

    def _get_combined_coverage_for_single_run(
        self,
        indicator_names: list[str],
        run: str | None = None,
        ensemble_numbers: list[int] | None = None,
        heights: list[int] | None = None,
        pressures: list[int] | None = None,
        intervals: list[str | None] | None = None,
        lat: tuple = FRANCE_METRO_LATITUDES,
        long: tuple = FRANCE_METRO_LONGITUDES,
        forecast_horizons: list[dt.timedelta] | None = None,
        temp_dir: str | None = None,
    ) -> pd.DataFrame:
        """(Protected)
        Get a combined DataFrame of coverage data for a given run considering a list of indicators.

        This method retrieves and aggregates coverage data for specified indicators, with options
        to filter by height, pressure, and forecast_horizon. It returns a concatenated DataFrame
        containing the coverage data.

        Args:
            indicator_names (list[str]): A list of indicator names to retrieve data for.
            run (str): A single runs for each indicator. Format should be "YYYY-MM-DDTHH:MM:SSZ".
            ensemble_numbers: For ensemble models only, numbers of the desired
                   ensemble members. If None, defaults to the member 0.
            heights (list[int] | None): A list of heights in meters to filter by (default is None).
            pressures (list[int] | None): A list of pressures in hPa to filter by (default is None).
            intervals (Optional[list[str]]): A list of aggregation periods (default is None).
                    Must be `None` or "" for instant indicators ; otherwise, raises an exception.
                    Defaults to 'P1D' for time-aggregated indicators.
            lat (tuple): The latitude range as (min_latitude, max_latitude). Defaults to FRANCE_METRO_LATITUDES.
            long (tuple): The longitude range as (min_longitude, max_longitude). Defaults to FRANCE_METRO_LONGITUDES.
            forecast_horizons (list[dt.timedelta] | None): A list of forecast horizon values (as a dt.timedelta object). Defaults to None.
            temp_dir (str | None): Directory to store the temporary file. Defaults to None.

        Returns:
            pd.DataFrame: A combined DataFrame containing coverage data for all specified runs and indicators.

        Raises:
            ValueError: If the length of `heights` does not match the length of `indicator_names`.
        """

        def _check_params_length(params: list[Any] | None, arg_name: str) -> list[Any]:
            """(Protected)
            Assert length is ok or raise an error.

            Args:
                params: list of parameters.
                arg_name: argument name.

            Returns:
                The given parameters unchanged.

            Raises:
                ValueError: The length of {arg_name} must match the length of indicator_names.
            """
            if params is None:
                return [None] * len(indicator_names)
            if len(params) != len(indicator_names):
                raise ValueError(
                    f"The length of {arg_name} must match the length of indicator_names."
                    f" If you want multiple {arg_name} for a single indicator, create multiple"
                    " entries in `indicator_names`."
                )
            return params

        heights = _check_params_length(heights, "heights")
        pressures = _check_params_length(pressures, "pressures")
        intervals = _check_params_length(intervals, "intervals")

        # Get coverage id from run and indicator_name
        coverage_ids = [
            self._get_coverage_id(indicator_name, run, interval)
            for indicator_name, interval in zip(indicator_names, intervals)
        ]

        if forecast_horizons:
            # Check forecast_horizons is valid for all indicators
            invalid_coverage_ids = self._validate_forecast_horizons(coverage_ids, forecast_horizons)
            if invalid_coverage_ids:
                raise ValueError(f"{forecast_horizons} are not valid for these coverage_ids : {invalid_coverage_ids}")
        else:
            forecast_horizons = [self.find_common_forecast_horizons(coverage_ids)[0]]
            logger.info(f"Using common forecast_horizons `forecast_horizons={forecast_horizons}`.")

        coverages = [
            [
                self.get_coverage(
                    coverage_id=coverage_id,
                    run=run,
                    lat=lat,
                    long=long,
                    ensemble_numbers=[ensemble_number] if ensemble_number is not None else None,
                    heights=[height] if height is not None else [],
                    pressures=[pressure] if pressure is not None else [],
                    forecast_horizons=forecast_horizons,
                    temp_dir=temp_dir,
                )
                for coverage_id, height, pressure in zip(coverage_ids, heights, pressures)
            ]
            for ensemble_number in ([None] if (ensemble_numbers is None) else ensemble_numbers)
        ]

        coverages_concat = pd.concat(
            [
                reduce(
                    lambda left, right: pd.merge(
                        left,
                        right,
                        on=["latitude", "longitude", "ensemble_number", "run", "forecast_horizon"]
                        if self.MODEL_TYPE == "ENSEMBLE"
                        else ["latitude", "longitude", "run", "forecast_horizon"],
                        how="inner",
                        validate="one_to_one",
                    ),
                    coverages[i],
                )
                for i in range(len(coverages))
            ]
        )

        return coverages_concat

    def _get_forecast_horizons(self, coverage_ids: list[str]) -> list[list[dt.timedelta]]:
        """(Protected)
        Retrieve the times for each coverage_id.

        Args:
            coverage_ids: List of coverage IDs.

        Returns:
            List of times for each coverage ID.
        """
        indicator_times: list[list[dt.timedelta]] = []
        for coverage_id in coverage_ids:
            times = self.get_coverage_description(coverage_id)["forecast_horizons"]
            indicator_times.append(times)
        return indicator_times

    def find_common_forecast_horizons(
        self,
        list_coverage_id: list[str],
    ) -> list[dt.timedelta]:
        """Find common forecast_horizons among coverage IDs.

        Args:
            indicator_names: List of indicator names.
            run: Identifies the model inference. Defaults to latest if None. Format "YYYY-MM-DDTHH:MM:SSZ".
            intervals: List of aggregation periods. Must be None for instant indicators, otherwise raises.
                    Defaults to P1D for time-aggregated indicators like TOTAL_PRECIPITATION.

        Returns:
            List of common forecast_horizons.
        """
        indicator_forecast_horizons = self._get_forecast_horizons(list_coverage_id)

        common_forecast_horizons = indicator_forecast_horizons[0]
        for times in indicator_forecast_horizons[1:]:
            common_forecast_horizons = [time for time in common_forecast_horizons if time in times]

        all_times = []
        for times in indicator_forecast_horizons:
            all_times.extend(times)

        return sorted(common_forecast_horizons)

    def _validate_forecast_horizons(self, coverage_ids: list[str], forecast_horizons: list[dt.timedelta]) -> list[str]:
        """(Protected)
        Validate forecast_horizons for a list of coverage IDs.

        Args:
            coverage_ids: List of coverage IDs.
            forecast_horizons: List of time forecasts to validate.

        Returns:
            List of invalid coverage IDs.
        """
        indicator_forecast_horizons = self._get_forecast_horizons(coverage_ids)

        invalid_coverage_ids = [
            coverage_id
            for coverage_id, times in zip(coverage_ids, indicator_forecast_horizons)
            if not set(forecast_horizons).issubset(times)
        ]

        return invalid_coverage_ids



from __future__ import annotations

import logging
from typing import final

from meteole.clients import BaseClient, MeteoFranceClient
from meteole.forecast import WeatherForecast

logger = logging.getLogger(__name__)

AVAILABLE_AROME_TERRITORY: list[str] = [
    "FRANCE",
]


@final
class AromePIForecast(WeatherForecast):
    """Access the AROME numerical weather forecast data from Meteo-France API.

    Doc:
        - https://portail-api.meteofrance.fr/web/fr/api/arome

    Attributes:
        territory: Covered area (e.g., FRANCE, ANTIL, ...).
        precision: Precision value of the forecast.
        capabilities: DataFrame containing details on all available coverage ids.
    """

    # Model constants
    MODEL_NAME: str = "aromepi"
    BASE_ENTRY_POINT: str = "wcs/MF-NWP-HIGHRES-AROMEPI"
    MODEL_TYPE: str = "DETER"
    ENSEMBLE_NUMBERS: int = 1
    DEFAULT_TERRITORY: str = "FRANCE"
    DEFAULT_PRECISION: float = 0.01
    CLIENT_CLASS: type[BaseClient] = MeteoFranceClient

    def _validate_parameters(self) -> None:
        """Check the territory and the precision parameters.

        Raise:
            ValueError: At least, one parameter is not good.
        """
        if self.precision not in [0.01, 0.025]:
            raise ValueError("Parameter `precision` must be in (0.01, 0.025). It is inferred from argument `territory`")

        if self.territory not in AVAILABLE_AROME_TERRITORY:
            raise ValueError(f"Parameter `territory` must be in {AVAILABLE_AROME_TERRITORY}")



from unittest.mock import MagicMock, patch

import pytest

from meteole.clients import MeteoFranceClient


def test_init_with_api_key():
    api = MeteoFranceClient(api_key="dummy_api_key")
    assert api._api_key == "dummy_api_key"
    assert api._token is None
    assert api._application_id is None


def test_init_with_token():
    api = MeteoFranceClient(token="dummy_token")
    assert api._api_key is None
    assert api._token == "dummy_token"
    assert api._application_id is None


@patch("meteole.clients.MeteoFranceClient._connect")
def test_init_with_application_id(mock_connect):
    api = MeteoFranceClient(application_id="dummy_app_id")
    assert api._application_id == "dummy_app_id"
    assert api._api_key is None


def test_connect_no_credentials():
    with pytest.raises(ValueError):
        MeteoFranceClient()


@patch("requests.post")
def test_get_token(mock_post):
    mock_post.return_value.json.return_value = {"access_token": "dummy_token"}

    api = MeteoFranceClient(application_id="dummy_app_id")
    token = api._get_token()

    assert token == "dummy_token"
    assert api._token == "dummy_token"


@patch("requests.Session.get")
@patch.object(MeteoFranceClient, "_get_token")
def test_get_request_success(mock_get_token, mock_get):
    api = MeteoFranceClient(api_key="dummy_api_key")
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"data": "some data"}
    mock_get.return_value = mock_response

    response = api.get("DUMMY_PATH")
    assert response.status_code == 200
    assert response.json() == {"data": "some data"}


@patch("requests.Session.get")
@patch.object(MeteoFranceClient, "_get_token")
def test_get_request_token_expired(mock_get_token, mock_get):
    api = MeteoFranceClient(api_key="dummy_api_key")
    api.RETRY_DELAY_SEC = 0

    expired_response = MagicMock()
    expired_response.status_code = 401
    expired_response.headers = {"Content-Type": "application/json"}
    expired_response.json.return_value = {"code": "900901"}

    valid_response = MagicMock()
    valid_response.status_code = 200
    valid_response.json.return_value = {"data": "some data"}

    mock_get.side_effect = [expired_response, valid_response, valid_response]

    response = api.get("DUMMY_PATH")
    assert response.status_code == 200
    assert response.json() == {"data": "some data"}

    response = api.get("DUMMY_PATH")
    assert response.status_code == 200
    assert response.json() == {"data": "some data"}


def test_token_expired():
    api = MeteoFranceClient(api_key="dummy_api_key")
    expired_response = MagicMock()
    expired_response.status_code = 401
    expired_response.headers = {"Content-Type": "application/json"}
    expired_response.json = lambda: {"code": "900901"}

    assert api._is_token_expired(expired_response) is True


def test_token_not_expired():
    api = MeteoFranceClient(api_key="dummy_api_key")
    valid_response = MagicMock()
    valid_response.status_code = 200
    valid_response.headers = {"Content-Type": "application/json"}
    valid_response.text = lambda: {"code": "900901"}

    assert api._is_token_expired(valid_response) == False


@patch("requests.Session.get")
@patch.object(MeteoFranceClient, "_get_token")
def test_get_request_specific_error(mock_get_token, mock_get):
    api = MeteoFranceClient(api_key="dummy_api_key")
    api.RETRY_DELAY_SEC = 0

    error_response = MagicMock()
    error_response.status_code = 502
    error_response.json.return_value = {"error": "Bad Gateway"}

    valid_response = MagicMock()
    valid_response.status_code = 200
    valid_response.json.return_value = {"data": "some data"}

    mock_get.side_effect = [error_response, valid_response]

    response = api.get("DUMMY_PATH")
    assert response.status_code == 200
    assert response.json() == {"data": "some data"}
    assert mock_get.call_count == 2



from __future__ import annotations

import logging
from typing import final

from meteole.clients import BaseClient, MeteoFranceClient
from meteole.forecast import WeatherForecast

logger = logging.getLogger(__name__)

AVAILABLE_AROME_TERRITORY: list[str] = [
    "FRANCE",
    "NCALED",
    "INDIEN",
    "POLYN",
    "GUYANE",
    "ANTIL",
]


@final
class AromePEForecast(WeatherForecast):
    """Access the PE-AROME ensemble forecast data from Meteo-France API."""

    MODEL_NAME: str = "pearome"
    BASE_ENTRY_POINT: str = "wcs/MF-NWP-HIGHRES-PEARO"
    MODEL_TYPE: str = "ENSEMBLE"
    ENSEMBLE_NUMBERS: int = 25
    DEFAULT_TERRITORY: str = "FRANCE"
    CLIENT_CLASS: type[BaseClient] = MeteoFranceClient

    def __init__(self, client=None, **kwargs):
        super().__init__(client, precision=0.025, **kwargs)

    def _validate_parameters(self) -> None:
        """Check the territory and the precision parameters.

        Raise:
            ValueError: At least, one parameter is not good.
        """
        if self.precision != 0.025:
            raise ValueError("Parameter `precision` must be 0.025")

        if self.territory not in AVAILABLE_AROME_TERRITORY:
            raise ValueError(f"Parameter `territory` must be in {AVAILABLE_AROME_TERRITORY}")



import unittest
from unittest.mock import MagicMock, patch

import pandas as pd

from meteole._vigilance import Vigilance
from meteole.clients import MeteoFranceClient


class TestVigilance(unittest.TestCase):
    def setUp(self):
        self.api_key = "fake_api_key"
        self.token = "fake_token"
        self.application_id = "fake_app_id"
        client = MeteoFranceClient(api_key=self.api_key, token=self.token, application_id=self.application_id)
        self.vigilance = Vigilance(client)

    @patch("meteole.clients.MeteoFranceClient.get")
    def test_get_vigilance_bulletin(self, mock_get_request):
        mock_response = MagicMock()
        mock_response.json.return_value = {"data": "some_data"}
        mock_get_request.return_value = mock_response

        result = self.vigilance.get_bulletin()
        self.assertEqual(result, {"data": "some_data"})
        # mock_get_request.assert_called_once_with(
        #     "https://public-api.meteofrance.fr/public/DPVigilance/v1/textesvigilance/encours"
        # )

    @patch("meteole.clients.MeteoFranceClient.get")
    def test_get_vigilance_map(self, mock_get_request):
        mock_response = MagicMock()
        mock_response.json.return_value = {"data": "some_data"}
        mock_get_request.return_value = mock_response

        result = self.vigilance.get_map()
        self.assertEqual(result, {"data": "some_data"})
        # mock_get_request.assert_called_once_with(
        #     "https://public-api.meteofrance.fr/public/DPVigilance/v1/cartevigilance/encours"
        # )

    @patch("meteole._vigilance.Vigilance.get_map")
    def test_get_phenomenon(self, mock_get_vigilance_map):
        mock_get_vigilance_map.return_value = {
            "product": {
                "periods": [
                    {
                        "echeance": "J",
                        "per_phenomenon_items": [
                            {
                                "phenomenon_id": "1",
                                "any_color_count": 5,
                                "phenomenon_counts": [{"color_id": 2, "color_name": "Jaune", "count": 3}],
                            }
                        ],
                        "timelaps.domain_ids": [{"domain_id": 1, "max_color_id": 1}],
                    },
                    {
                        "echeance": "J1",
                        "per_phenomenon_items": [
                            {
                                "phenomenon_id": "2",
                                "any_color_count": 5,
                                "phenomenon_counts": [{"color_id": 2, "color_name": "Jaune", "count": 3}],
                            }
                        ],
                        "timelaps.domain_ids": [{"domain_id": 2, "max_color_id": 2}],
                    },
                ]
            }
        }

        with patch.object(self.vigilance, "get_map", return_value=mock_get_vigilance_map.return_value):
            df_phenomenon, df_timelaps = self.vigilance.get_phenomenon()

            expected_phenomenon_data = {
                "phenomenon_id": ["1", "2"],
                "any_color_count": [5, 5],
                "phenomenon_counts": [
                    [{"color_id": 2, "color_name": "Jaune", "count": 3}],
                    [{"color_id": 2, "color_name": "Jaune", "count": 3}],
                ],
                "echeance": ["J", "J1"],
                "phenomenon_libelle": ["wind", "rain"],
            }
            expected_phenomenon_df = pd.DataFrame(expected_phenomenon_data)

            expected_timelaps_data = {
                "domain_id": [1, 2],
                "max_color_id": [1, 2],
                "echeance": ["J", "J1"],
            }
            expected_timelaps_df = pd.DataFrame(expected_timelaps_data)

            pd.testing.assert_frame_equal(df_phenomenon, expected_phenomenon_df)
            pd.testing.assert_frame_equal(df_timelaps, expected_timelaps_df)



from importlib.metadata import version

from meteole._arome import AromeForecast
from meteole._arome_ensemble import AromePEForecast
from meteole._arome_instantane import AromePIForecast
from meteole._arpege import ArpegeForecast
from meteole._piaf import PiafForecast
from meteole._vigilance import Vigilance

__all__ = ["AromeForecast", "AromePIForecast", "ArpegeForecast", "PiafForecast", "Vigilance", "AromePEForecast"]

__version__ = version("meteole")



"""Error implementations"""

from __future__ import annotations

from typing import Any, MutableMapping

import xmltodict


class GenericMeteofranceApiError(Exception):
    """Exception raised when a required field is missing in the input parameters.

    Args:
        message: Human-readable string describing the exception.
        description: More detailed description of the error.
    """

    def __init__(self, text: str) -> None:
        """Initialize the exception with an error message parsed from an XML
        string.

        Args:
            text: XML string containing the error details,
                expected to follow a specific schema with 'am:fault' as the root
                element and 'am:message' and 'am:description' as child elements."""
        try:
            # Parse error message with xmltodict
            data: MutableMapping[str, Any] = xmltodict.parse(text)
            msg_content: str = data["am:fault"]["am:message"]
            description: str = data["am:fault"]["am:description"]
            message: str = f"{msg_content}\n {description}"
        except Exception:
            message = text

        super().__init__(message)


class MissingDataError(Exception):
    """Exception raised errors in the input data is missing"""

    def __init__(self, text: str) -> None:
        """Initialize the exception with an error message parsed from an XML
        string.

        Args:
            text: XML string containing the error details,
                expected to follow a specific schema with 'am:fault' as the root
                element and 'am:message' and 'am:description' as child elements."""

        try:
            # Parse error message with xmltodict
            data: MutableMapping[str, Any] = xmltodict.parse(text)
            exception: dict[Any, Any] = data["mw:fault"]["mw:description"]["ns0:ExceptionReport"]["ns0:Exception"]
            code: str = exception["@exceptionCode"]
            locator: str = exception["@locator"]
            exception_text: str = exception["ns0:ExceptionText"]
            message: str = f"Error code: {code}\nLocator: {locator}\nText: {exception_text}"
        except Exception:
            message = text

        super().__init__(message)



"""API client classes"""

from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any

import requests
from requests import Response, Session

from meteole.errors import GenericMeteofranceApiError, MissingDataError

logger = logging.getLogger(__name__)


class HttpStatus(int, Enum):
    """Http status codes"""

    OK = 200
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_FOUND = 404
    TOO_MANY_REQUESTS = 429
    INTERNAL_ERROR = 500
    BAD_GATEWAY = 502
    UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504


class BaseClient(ABC):
    """(Abstract)

    Base class for weather forecast provider clients.
    """

    @abstractmethod
    def get(self, path: str, *, params: dict[str, Any] | None = None, max_retries: int = 5) -> Response:
        """Retrieve some data with retry capability.

        Args:
            path: Path to a resource.
            params: The query parameters of the request.
            max_retries: The maximum number of retry attempts in case of failure.

        Returns:
            The response returned by the API.
        """
        raise NotImplementedError


class MeteoFranceClient(BaseClient):
    """A client for interacting with the Meteo France API.

    This class handles the connection setup and token refreshment required for
    authenticating and making requests to the Meteo France API.
    """

    # Class constants
    TOKEN_URL: str = "https://portail-api.meteofrance.fr/token"  # noqa: S105
    GET_TOKEN_TIMEOUT_SEC: int = 10
    INVALID_JWT_ERROR_CODE: str = "900901"
    RETRY_DELAY_SEC: int = 10

    def __init__(
        self,
        *,
        token: str | None = None,
        api_base_url: str = "https://public-api.meteofrance.fr/public/",  # need it as an argument since PIAF model has a different base URL
        api_key: str | None = None,
        application_id: str | None = None,
        certs_path: Path | None = None,
    ) -> None:
        """
        Initialize attributes.

        Args:
            token: The authentication token for accessing the API.
            api_key: The API key for accessing the Meteo France API.
            application_id: The application ID used for identification.
            certs_path: The path to a file or directory of trusted CA certificates for SSL verification.
        """
        self._api_base_url = api_base_url
        self._token = token
        self._api_key = api_key
        self._application_id = application_id
        self._verify: str | None = str(certs_path) if certs_path is not None else None

        self._session = Session()

        self._token_expired: bool = False

        # Initialize the requests session object
        self._connect()

    def get(self, path: str, *, params: dict[str, Any] | None = None, max_retries: int = 5) -> Response:
        """
        Make a GET request to the API with optional retries.

        Args:
            path: Path to a resource.
            params: The query parameters of the request.
            max_retries: The maximum number of retry attempts in case of failure.

        Returns:
            The response returned by the API.
        """
        url: str = self._api_base_url + path
        attempt: int = 0
        logger.debug(f"GET {url}")

        while attempt < max_retries:
            # HTTP GET request
            try:
                resp: Response = self._session.get(url, params=params, verify=self._verify)

                if resp.status_code == HttpStatus.OK:
                    logger.debug("Successful request")
                    return resp

                elif self._is_token_expired(resp):
                    logger.info("Token expired, requesting a new one")

                    # Refresh the cached token
                    self._token = self._get_token()

                    # Reconnect with the new token
                    self._connect()

                elif resp.status_code == HttpStatus.FORBIDDEN:
                    logger.error("Access forbidden")
                    raise GenericMeteofranceApiError(resp.text)

                elif resp.status_code == HttpStatus.BAD_REQUEST:
                    logger.error("Parameter error")
                    raise GenericMeteofranceApiError(resp.text)

                elif resp.status_code == HttpStatus.NOT_FOUND:
                    logger.error("Missing data")
                    raise MissingDataError(resp.text)

                elif (
                    resp.status_code == HttpStatus.BAD_GATEWAY
                    or resp.status_code == HttpStatus.UNAVAILABLE
                    or resp.status_code == HttpStatus.GATEWAY_TIMEOUT
                ):
                    logger.error("Service not available")

            except requests.exceptions.ConnectionError as e:
                logger.warning(f"Connection error : {e}.")

            # Wait before retrying
            attempt += 1
            waiting_time = attempt * self.RETRY_DELAY_SEC
            logger.info(f"Retrying (attempt {attempt}/{max_retries}) - waiting {waiting_time}s before retrying...")
            time.sleep(waiting_time)
            continue

        raise GenericMeteofranceApiError(f"Failed to get a successful response from API after {attempt} retries")

    def _connect(self):
        """(Protected)
        Connect to the Meteo-France API.

        Note:
        If the API key is provided, it is used to authenticate the user.
        If the token is provided, it is used to authenticate the user.
        If the application ID is provided, a token is requested from the API.
        """
        if self._api_key is None and self._token is None:
            if self._application_id is None:
                raise ValueError("api_key or token or application_id must be provided")

            # Connection with application_id
            self._token = self._get_token()

        if self._api_key is not None:
            logger.debug("using api key")

            # Connection with api_key
            self._session.headers.update({"apikey": self._api_key})

        else:
            logger.debug("using token")

            # Connection with token
            self._session.headers.update({"Authorization": f"Bearer {self._token}"})

    def _get_token(self) -> str:
        """(Protected)
        Request a token from the Meteo-France API.

        The token lasts 1 hour, and is used to authenticate the user.
        If a new token is requested before the previous one expires, the previous one is invalidated.
        A local cache is used to avoid requesting a new token at each run of the script.

        Rerturns:
            A JWT.
        """
        if self._token_expired is False and self._token is not None:
            # Use cached token
            token: str = self._token

        elif self._token_expired is True and self._application_id is None:
            # Can do nothing
            raise ValueError("The 'application_id' is unknown, can't get a new token")

        else:
            # Retrieve a new token

            # Seems useless (TODO remove if it's True):
            # params: dict[str, str] = {"grant_type": "client_credentials"}
            headers: dict[str, str] = {"Authorization": "Basic " + str(self._application_id)}

            resp: Response = requests.post(
                self.TOKEN_URL,
                # params=params,
                headers=headers,
                timeout=self.GET_TOKEN_TIMEOUT_SEC,
                verify=self._verify,
            )
            token = resp.json()["access_token"]

        return token

    def _is_token_expired(self, response: Response) -> bool:
        """(Protected)
        Check if the token is expired.

        Args:
            response: A request's response from the API.

        Returns:
            True if the token is expired, False otherwise.
        """
        result: bool = False

        if response.status_code == HttpStatus.UNAUTHORIZED and "application/json" in response.headers["Content-Type"]:
            error: str = response.json()["code"]

            if error == self.INVALID_JWT_ERROR_CODE:
                result = True
                self._token_expired = True

        return result



from __future__ import annotations

import logging
from typing import Any, final

from meteole.clients import BaseClient, MeteoFranceClient
from meteole.forecast import WeatherForecast

logger = logging.getLogger(__name__)

AVAILABLE_PIAF_TERRITORY: list[str] = [
    "FRANCE",
]


@final
class PiafForecast(WeatherForecast):
    """Access the PIAF numerical weather forecast data from Meteo-France API.

    Doc:
        - https://portail-api.meteofrance.fr/web/fr/api/arome

    Attributes:
        territory: Covered area (e.g., FRANCE, ANTIL, ...).
        precision: Precision value of the forecast.
        capabilities: DataFrame containing details on all available coverage ids.
    """

    # Model constants
    MODEL_NAME: str = "piaf"
    BASE_ENTRY_POINT: str = "wcs/MF-NWP-HIGHRES-PIAF"
    MODEL_TYPE: str = "DETER"
    ENSEMBLE_NUMBERS: int = 1
    DEFAULT_TERRITORY: str = "FRANCE"
    DEFAULT_PRECISION: float = 0.01
    CLIENT_CLASS: type[BaseClient] = MeteoFranceClient

    def __init__(
        self,
        **kwargs: Any,
    ):
        """Initialize attributes.

        Args:
            api_key: The API key for authentication. Defaults to None.
            token: The API token for authentication. Defaults to None.
            application_id: The Application ID for authentication. Defaults to None.
        """
        super().__init__(api_base_url="https://api.meteofrance.fr/pro/", **kwargs)

    def _validate_parameters(self) -> None:
        """Check the territory and the precision parameters.

        Raise:
            ValueError: At least, one parameter is not good.
        """
        if self.precision != 0.01:
            raise ValueError("Parameter `precision` must be in 0.01.")

        if self.territory not in AVAILABLE_PIAF_TERRITORY:
            raise ValueError(f"Parameter `territory` must be in {AVAILABLE_PIAF_TERRITORY}")
